# -*- coding: utf-8 -*-
"""
本模块功能：上市公司的财务报表分析，数据层
所属工具包：证券投资分析工具SIAT 
SIAT：Security Investment Analysis Tool
创建日期：2020年9月10日
最新修订日期：2020年9月14日
作者：王德宏 (WANG Dehong, Peter)
作者单位：北京外国语大学国际商学院
版权所有：王德宏
用途限制：仅限研究与教学使用，不可商用！商用需要额外授权。
特别声明：作者不对使用本工具进行证券投资导致的任何损益负责！
"""

#==============================================================================
#关闭所有警告
import warnings; warnings.filterwarnings('ignore')
#==============================================================================
if __name__=='__main__':
    symbol='INTL' 
    symbol='600519.SS'

def get_balance_sheet(symbol):
    """
    功能：获取雅虎财经上一只股票所有的年度和季度资产负债表
    """
    print("...Searching for balance info of",symbol,"\b, please wait...",end='')
    
    from yahooquery import Ticker
    stock = Ticker(symbol)    
    
    import time
    #获取年报
    try:
        stmta=stock.balance_sheet()  # Defaults to Annual
    except:
        print("\nSearching for annual report failed, recovering......")
        time.sleep(5)
        try:
            stmta=stock.balance_sheet()
        except:
            print("#Error(get_balance_sheet): no annual info available for",symbol)
            return None
    
    #判断是否抓取到了数据
    import pandas as pd
    if not isinstance(stmta,pd.DataFrame):
        print("\n#Error(get_balance_sheet): no annual info available for",symbol)
        return None
    
    #获取季度报
    try:
        stmtq=stock.balance_sheet(frequency="q")
    except:
        print("Searching for quarterly report failed, recovering......")
        time.sleep(5)
        try:
            stmtq=stock.balance_sheet(frequency="q")
        except:
            print("#Error(get_balance_sheet): no quarterly info available for",symbol)
            return None

    if (len(stmta)==0) and (len(stmtq)==0):
        print("#Error(get_balance_sheet): no reports available for",symbol)
        return None

    #合并年度和季度报表
    import pandas as pd
    stmt=pd.concat([stmta,stmtq])
    
    #字段异名处理
    if 'CurrentDebt' not in list(stmt):
        stmt['CurrentDebt']=stmt['CurrentLiabilities']
    if 'TotalDebt' not in list(stmt):
        stmt['TotalDebt']=stmt['TotalLiabilitiesNetMinorityInterest']    
    
    #排序
    stmt.sort_values(by=['asOfDate','periodType'],inplace=True)
    #去掉重复记录: 保留年报数据项多，去掉数据项少的季报
    stmt.drop_duplicates(subset=['asOfDate'],keep='first',inplace=True)

    print(", done")
    return stmt    
    
if __name__ == '__main__':
    fs=get_balance_sheet('AAPL')
    fs_msft=get_balance_sheet('MSFT')
    fs_c=get_balance_sheet('C')
    fs_maotai=get_balance_sheet('600519.SS')

#==============================================================================
if __name__=='__main__':
    symbol='AAPL' 
    symbol='MST'

def get_income_statements(symbol):
    """
    功能：获取雅虎财经上一只股票所有的年度和季度利润表
    """
    print("...Searching for income info of",symbol,"\b, please wait...",end='')
    
    from yahooquery import Ticker
    stock = Ticker(symbol)    
    
    import time
    #获取年报
    try:
        stmta=stock.income_statement()  # Defaults to Annual
    except:
        print("\nSearching for annual report failed, recovering......")
        time.sleep(5)
        try:
            stmta=stock.income_statement()
        except:
            print("#Error(get_income_statements): no annual info available for",symbol)
            return None
    
    #判断是否抓取到了数据
    import pandas as pd
    if not isinstance(stmta,pd.DataFrame):
        print("\n#Error(get_income_statements): no annual info available for",symbol)
        return None
    
    #获取季度报
    try:
        stmtq=stock.income_statement(frequency="q")
    except:
        print("Searching for quarterly report failed, recovering......")
        time.sleep(5)
        try:
            stmtq=stock.income_statement(frequency="q")
        except:
            print("#Error(get_income_statements): no quarterly info available for",symbol)
            return None

    if (len(stmta)==0) and (len(stmtq)==0):
        print("#Error(get_income_statements): no reports available for",symbol)
        return None

    #合并年度和季度报表
    import pandas as pd
    stmt=pd.concat([stmta,stmtq])
    
    #排序
    stmt.sort_values(by=['asOfDate','periodType'],inplace=True)
    #去掉重复记录: 保留年报数据项多，去掉数据项少的季报
    stmt.drop_duplicates(subset=['asOfDate'],keep='first',inplace=True)

    #字段缺失处理
    if 'InterestExpense' not in list(stmt):
        import numpy as np
        stmt['InterestExpense']=np.nan
    
    print(", done")
    return stmt    
    
if __name__ == '__main__':
    stmt=get_income_statements('AAPL')
    stmt_msft=get_income_statements('MSFT')
    stmt_c=get_income_statements('C')
    stmt_maotai=get_income_statements('600519.SS')

#==============================================================================
if __name__=='__main__':
    symbol='AAPL' 

def get_cashflow_statements(symbol):
    """
    功能：获取雅虎财经上一只股票所有的年度和季度现金流量表
    """
    print("...Searching for cashflow info of",symbol,"\b, please wait...",end='')
    
    from yahooquery import Ticker
    stock = Ticker(symbol)    
    
    import time
    #获取年报
    try:
        stmta=stock.cash_flow()  # Defaults to Annual
    except:
        print("\nSearching for annual report failed, recovering......")
        time.sleep(5)
        try:
            stmta=stock.cash_flow()
        except:
            print("#Error(get_cashflow_statements): no annual info available for",symbol)
            return None
    
    #判断是否抓取到了数据
    import pandas as pd
    if not isinstance(stmta,pd.DataFrame):
        print("\n#Error(get_cashflow_statements): no annual info available for",symbol)
        return None
    
    #获取季度报
    try:
        stmtq=stock.cash_flow(frequency="q")
    except:
        print("Searching for quarterly report failed, recovering......")
        time.sleep(5)
        try:
            stmtq=stock.cash_flow(frequency="q")
        except:
            print("#Error(get_cashflow_statements): no quarterly info available for",symbol)
            return None

    if (len(stmta)==0) and (len(stmtq)==0):
        print("#Error(get_cashflow_statements): no reports available for",symbol)
        return None

    #合并年度和季度报表
    import pandas as pd
    stmt=pd.concat([stmta,stmtq])
    
    #排序
    stmt.sort_values(by=['asOfDate','periodType'],inplace=True)
    #去掉重复记录: 保留年报数据项多，去掉数据项少的季报
    stmt.drop_duplicates(subset=['asOfDate'],keep='first',inplace=True)

    #字段缺失处理
    if 'CashDividendsPaid' not in list(stmt):
        import numpy as np
        stmt['CashDividendsPaid']=np.nan

    print(", done")
    return stmt    
    
if __name__ == '__main__':
    stmt=get_cashflow_statements('AAPL')
    stmt_msft=get_cashflow_statements('MSFT')
    stmt_c=get_cashflow_statements('C')
    stmt_maotai=get_cashflow_statements('600519.SS')

#==============================================================================
if __name__=='__main__':
    ticker='AAPL' 

def get_financial_statements(ticker):
    """
    功能：获取雅虎财经上一只股票所有的年度和季度财务报表
    """
    print("Searching for financial statements of",ticker,"\b, please wait")
    #获取资产负债表
    try:
        fbs = get_balance_sheet(ticker)
    except:
        print("      Failed, recovering...")
        import time; time.sleep(3)
        try:
            fbs = get_balance_sheet(ticker)
        except:
            import time; time.sleep(5)
            try:
                fbs = get_balance_sheet(ticker)
            except:
                print("#Error(get_financial_statements): balance sheet not available for",ticker)
                return None
    if fbs is None:
        print("#Error(get_financial_statements): financial statements not available for",ticker)
        return None
    
    #获取利润表
    try:
        fis = get_income_statements(ticker)
    except:
        print("      Failed, recovering...")
        import time; time.sleep(3)
        try:
            fis = get_income_statements(ticker)
        except:
            import time; time.sleep(5)
            try:
                fis = get_income_statements(ticker)
            except:
                print("#Error(get_financial_statements): income info not available for",ticker)
                return None
    if fis is None:
        print("#Error(get_financial_statements): financial statements not available for",ticker)
        return None
    
    #获取现金流量表
    try:
        fcf = get_cashflow_statements(ticker)
    except:
        print("      Failed, recovering...")
        import time; time.sleep(3)
        try:
            fcf = get_cashflow_statements(ticker)
        except:
            import time; time.sleep(5)
            try:
                fcf = get_cashflow_statements(ticker)
            except:
                print("#Error(get_financial_statements): cash flow info not available for",ticker)
                return None
    if fcf is None:
        print("#Error(get_financial_statements): financial statements not available for",ticker)
        return None
    
    #合并1：资产负债表+利润表
    import pandas as pd
    fbs_fis=pd.merge(fbs,fis,on=['asOfDate','periodType'])
    
    #合并2：+现金流量表
    fbs_fis_fcf=pd.merge(fbs_fis,fcf,on=['asOfDate','periodType','NetIncome'])
    
    fbs_fis_fcf['ticker']=ticker

    print("...Successfully retrieved financial statements of",ticker,"\b!")    
    return fbs_fis_fcf    
    
if __name__ == '__main__':
    fs=get_financial_statements('AAPL')
    fs_aapl=get_financial_statements('AAPL')
    fs_msft=get_financial_statements('MSFT')
    fs_c=get_financial_statements('C')
    fs_maotai=get_financial_statements('600519.SS')

"""
最终获得的表结构：
['asOfDate',
 'periodType',
 
 'AccountsPayable(应付账款)',
 'AccountsReceivable(应收账款)',
 'AccumulatedDepreciation（累计折旧）',
 'AdditionalPaidInCapital（资本公积，资本溢价，附加资本；paid-in capital：实收资本；缴入资本）',
 'AllowanceForDoubtfulAccountsReceivable（备抵应收呆帐）',
 'AvailableForSaleSecurities（可供出售金融资产；trading securities: 交易性金融资产）',
 'BuildingsAndImprovements（建筑物改良）',
 'CapitalStock（股本）',
 'CashAndCashEquivalents（现金及现金等价物）',
 'CashCashEquivalentsAndShortTermInvestments（现金、现金等价物及短期投资）',
 'CashEquivalents（现金等价物）',
 'CashFinancial（？）',
 'CommonStock（普通股）',
 'CommonStockEquity（普通股权益？）',
 'ConstructionInProgress（在建工程）',
 'CurrentAssets（流动资产）',
 'CurrentLiabilities（流动负债）',
 'DividendsPayable（应付股利）',
 'FinishedGoods（制成品）',
 'GoodwillAndOtherIntangibleAssets（商誉及其他无形资产）',
 'GrossAccountsReceivable（应收账款总额）',
 'GrossPPE（固定资产总额）',
 'InventoriesAdjustmentsAllowances（存货调整备抵）',
 'Inventory（存货）',
 'InvestedCapital（投入资本）',
 'InvestmentinFinancialAssets（金融资产投资？）',
 'LandAndImprovements（土地改良）',
 'MachineryFurnitureEquipment（机械家具设备？）',
 'MinorityInterest（少数股东损益？）',
 'NetPPE（固定资产净值）',
 'NetTangibleAssets（有形资产净值）',
 'NonCurrentDeferredAssets（非流动递延资产）',
 'NonCurrentDeferredTaxesAssets（非流动递延税项资产？）',
 'NonCurrentDeferredTaxesLiabilities（非流动递延税金负债？）',
 'OrdinarySharesNumber（普通股数量？）',
 'OtherCurrentAssets（其他流动资产）',
 'OtherCurrentLiabilities（其他流动负债）',
 'OtherEquityInterest（其他股权）',
 'OtherIntangibleAssets（其他有形资产）',
 'OtherNonCurrentAssets（其他非流动资产）',
 'OtherPayable（其它应付款）',
 'OtherProperties（？）',
 'OtherReceivables（其他应收款）',
 'Payables（应付款项）',
 'PrepaidAssets（预付资产；预付费用）',
 'Properties（财产？）',
 'RawMaterials（原材料）',
 'RetainedEarnings（留存收益）',
 'ShareIssued（股票发行）',
 'StockholdersEquity（股东权益）',
 'TangibleBookValue（有形资产账面价值）',
 'TotalAssets（总资产）',
 'TotalCapitalization（资本总额？）',
 'TotalEquityGrossMinorityInterest（少数股东权益总额）',
 'TotalLiabilitiesNetMinorityInterest（？）',
 'TotalNonCurrentAssets（非流动资产总额）',
 'TotalNonCurrentLiabilitiesNetMinorityInterest（？）',
 'TotalTaxPayable（应缴税款总额）',
 'TradeandOtherPayablesNonCurrent（？）',
 'WorkInProcess（在制品）',
 'WorkingCapital（营运资本）',
 'Amortization（摊销）',
 
 'BasicAverageShares（未稀释的平均股数？）',
 'BasicEPS（ 基本每股收益，指属于普通股股东的当期净利润，除以发行在外普通股的加权平均数，可按存在月数加权）',
 'CostOfRevenue（主营业务成本，营收成本）',
 'DepreciationAndAmortizationInIncomeStatement（损益表中的折旧和摊销）',
 'DepreciationIncomeStatement（损益表中的折旧）',
 'DilutedAverageShares（稀释后平均股数？）',
 'DilutedEPS（考虑了可转换债券和股票期权可能行权对于流通在外股数的影响）',
 'EBIT（息税前利润）',
 'EBITDA（未计利息、税项、折旧及摊销前的利润）',
 'GeneralAndAdministrativeExpense（一般管理费用）',
 'GrossProfit（营业毛利）',
 'ImpairmentOfCapitalAssets（资本资产减值）',
 'InterestExpense（利息费用）',
 'InterestExpenseNonOperating（非经营性利息费用）',
 'InterestIncome（利息收益）',
 'InterestIncomeNonOperating（非经营性利息收入）',
 'MinorityInterests（少数股东权益）',
 'NetIncome（净利润）',
 'NetIncomeCommonStockholders（归属于普通股股东的净利润，用于计算EPS和PE）',
 'NetIncomeContinuousOperations（扣非后净利润）',
 'NetIncomeFromContinuingAndDiscontinuedOperation（来自持续经营和停止经营业务的净收入）',
 'NetIncomeFromContinuingOperationNetMinorityInterest（不归属少数股东的扣非后净利润？）',
 'NetIncomeIncludingNoncontrollingInterests（包括非控股权的净收入？）',
 'NetInterestIncome（净利息收入）',
 'NetNonOperatingInterestIncomeExpense（？）',
 'NormalizedEBITDA（调整后EBITDA？）',
 'NormalizedIncome（调整后利润？）',
 'OperatingExpense（营业费用）',
 'OperatingIncome（营业利润）',
 'OperatingRevenue（营业收入）',
 'OtherNonOperatingIncomeExpenses（其他营业外收入支出？）',
 'OtherOperatingExpenses（其它营业费用）',
 'OtherSpecialCharges（其他特殊费用）',
 'OtherunderPreferredStockDividend（优先股股利下的其他项目）',
 'PretaxIncome（税前利润）',
 'ReconciledCostOfRevenue（对账后的经营收入成本？）',
 'ReconciledDepreciation（对账后的折旧）',
 'RentAndLandingFees（租金及土地费用？）',
 'RentExpenseSupplemental（补充租金费用？）',
 'ResearchAndDevelopment（研发费用？）',
 'SellingAndMarketingExpense（销售和市场营销费用）',
 'SellingGeneralAndAdministration（销售及一般管理费用）',
 'SpecialIncomeCharges（特殊的收入费用？）',
 'TaxEffectOfUnusualItems（特殊项目的税收效应？）',
 'TaxProvision（税金？）',
 'TaxRateForCalcs（计算用的税率）',
 'TotalExpenses（总费用）',
 'TotalOperatingIncomeAsReported（报告的总营业利润）',
 'TotalOtherFinanceCost（其他财务成本合计）',
 'TotalRevenue（总收入）',
 'TotalUnusualItems（非经常性项目总计）',
 'TotalUnusualItemsExcludingGoodwill（不包括商誉的非经常项目合计）',
 'WriteOff（冲销？）',
 
 'BeginningCashPosition（期初现金头寸）',
 'CapitalExpenditure（资本支出）',
 'CashDividendsPaid（现金股利支付）',
 'ChangeInCashSupplementalAsReported（现金补充变更报告？）',
 'ChangeInInventory（存货变化）',
 'ChangeInWorkingCapital（营运资本的变动额）',
 'DepreciationAndAmortization（折旧摊销）',
 'EndCashPosition（期末现金头寸）',
 'FreeCashFlow（自有现金流）',
 'InvestingCashFlow（投资现金流）',
 'NetOtherFinancingCharges（其他融资费用净额）',
 'NetOtherInvestingChanges（其他投资变动净额）',
 'OperatingCashFlow（营运现金流）',
 'OtherNonCashItems（其他非现金项目）'
 ]
"""

#==============================================================================
def date_adjust(basedate, adjust=0):
    """
    功能：将给定日期向前或向后调整特定的天数
    输入：基础日期，需要调整的天数。
    basedate: 基础日期。
    adjust：需要调整的天数，负数表示向前调整，正数表示向后调整。
    输出：调整后的日期。
    """
    #检查基础日期的合理性
    import pandas as pd    
    try:
        bd=pd.to_datetime(basedate)
    except:
        print("#Error(date_adjust)，invalid date:",basedate)
        return None

    #调整日期
    from datetime import timedelta
    nd = bd+timedelta(days=adjust)    
    
    #重新提取日期
    newdate=nd.date()   
    return str(newdate)
 
if __name__ =="__main__":
    basedate='2020-3-17' 
    adjust=-365    
    newdate = date_adjust(basedate, adjust)
    print(newdate)    

#==============================================================================
if __name__ =="__main__":
    ticker='INTL'
    fromdate='2016-1-1'
    todate='2020-8-31' 

def get_price(ticker, fromdate, todate):
    """
    功能：从雅虎财经抓取单个证券价格或指数价格，不能处理多个证券列表。
    输入：证券代码，开始日期，结束日期。
    ticker: 证券代码。
    大陆股票代码加上后缀.SZ或.SS，港股代码去掉前导0加后缀.HK。
    fromdate: 样本开始日期。
    todate: 样本结束日期。既可以是今天日期，也可以是一个历史日期。    
    输出：证券价格序列，按照日期升序排列。原汁原味的抓取数据。
    """

    #仅支持单个证券代码
    if not isinstance(ticker,str):
        print("*** 错误#1(get_price)，仅支持单个证券代码!")        
        return None  
    
    #校验证券代码
    _,ticker=ticker_check(ticker, source="yahoo")

    #检查期间合理性
    result,start,end=check_period(fromdate,todate)
    if result is None:
        print("*** 错误#2(get_price)，无效的日期或期间!")        
        return None         
    
    #使用pandas_datareader抓取雅虎证券价格
    try:
        from pandas_datareader import data
    except:
        print("*** 错误#3：需要先安装插件pandas-datareader，然后重新运行！")
        print("安装方法：")
        print("打开Anaconda prompt，执行命令：pip install pandas-datareader")
        return None    
    
    try:
        price=data.DataReader(ticker,'yahoo',start,end)
    except:
        print("*** 错误#4(get_price)，抓取证券价格失败",ticker,fromdate,todate)        
        return None            
    
    #抓取未失败但返回空数据
    if len(price)==0:
        print("*** 错误#5(get_price)，未能抓取到有效价格信息!")
        return None         
    
    """
    #去掉比起始日期更早的样本
    price2=price[price.index >= fromdate]
    #去掉比结束日期更晚的样本
    price2=price2[price2.index <= todate]
    
    #按日期升序排序，近期的价格排在后面
    sortedprice=price2.sort_index(axis=0,ascending=True)
    """
    
    #生成字符串型的日期字段date，方便以后使用
    price['datetmp']=price.index
    price['date']=price['datetmp'].apply(lambda x:x.date())
    del price['datetmp']

    #加入星期几
    price["Weekday"]=price.index.weekday + 1  
    
    #加入证券代码，便于后期处理
    price['ticker']=ticker

    return price

if __name__ =="__main__":
    df=get_price('000002.SZ','2020-2-1','2020-3-31')
    df=get_price('^GSPC','2020-1-1','2020-3-4')
    print(df['Close'].tail(10))
    price=get_price(['000001.SS','^HSI'],'2020-1-1','2020-2-4')
    apclose=price['Close']['^HSI']
    print(price['Close']['^HSI'].tail(10))

#==============================================================================
def ticker_check(ticker, source="yahoo"):
    """
    检查证券代码，对于大陆证券代码、香港证券代码和东京证券代码进行修正。
    输入：证券代码ticker，数据来源source。
    上交所证券代码后缀为.SS或.SH或.ss或.sh，深交所证券代码为.SZ或.sz
    港交所证券代码后缀为.HK，截取数字代码后4位
    东京证交所证券代码后缀为.T，截取数字代码后4位
    source：yahoo或tushare
    返回：字母全部转为大写。若是大陆证券返回True否则返回False。
    若选择yahoo数据源，上交所证券代码转为.SS；
    若选择tushare数据源，上交所证券代码转为.SH
    """
    #测试用，完了需要注释掉
    #ticker="600519.sh"
    #source="yahoo"
    
    #将字母转为大写
    ticker1=ticker.upper()
    #截取字符串最后2/3位
    suffix2=ticker1[-2:]
    suffix3=ticker1[-3:]
    
    #判断是否大陆证券
    if suffix3 in ['.SH', '.SS', '.SZ']:
        prc=True
    else: prc=False

    #根据数据源的格式修正大陆证券代码
    if (source == "yahoo") and (suffix3 in ['.SH']):
        ticker1=ticker1.replace(suffix3, '.SS')        
    if (source == "tushare") and (suffix3 in ['.SS']):
        ticker1=ticker1.replace(suffix3, '.SH')  

    #若为港交所证券代码，进行预防性修正，截取数字代码后4位，加上后缀共7位
    if suffix3 in ['.HK']:
        ticker1=ticker1[-7:]     

    #若为东交所证券代码，进行预防性修正，截取数字代码后4位，加上后缀共6位
    if suffix2 in ['.T']:
        ticker1=ticker1[-6:]  
    
    #返回：是否大陆证券，基于数据源/交易所格式修正后的证券代码
    return prc, ticker1        

#测试各种情形
if __name__=='__main__':
    prc, ticker=ticker_check("600519.sh","yahoo")
    print(prc,ticker)
    print(ticker_check("600519.SH","yahoo"))    
    print(ticker_check("600519.ss","yahoo"))    
    print(ticker_check("600519.SH","tushare"))    
    print(ticker_check("600519.ss","tushare"))    
    print(ticker_check("000002.sz","tushare"))
    print(ticker_check("000002.sz","yahoo"))
    print(ticker_check("00700.Hk","yahoo"))
    print(ticker_check("99830.t","yahoo"))
#==============================================================================
def check_period(fromdate, todate):
    """
    功能：根据开始/结束日期检查日期与期间的合理性
    输入参数：
    fromdate：开始日期。格式：YYYY-MM-DD
    enddate：开始日期。格式：YYYY-MM-DD
    输出参数：
    validity：期间合理性。True-合理，False-不合理
    start：开始日期。格式：datetime类型
    end：结束日期。格式：datetime类型
    """
    import pandas as pd
    
    #测试开始日期的合理性
    try:
        start=pd.to_datetime(fromdate)
    except:
        print("*** 错误#1(check_period)，无效的日期:",fromdate)
        return None, None, None         
    
    #测试结束日期的合理性
    try:
        end=pd.to_datetime(todate)
    except:
        print("*** 错误#2(check_period)，无效的日期:",todate)
        return None, None, None          
    
    #测试日期期间的合理性
    if start > end:
        print("*** 错误#3(check_period)，无效的日期期间: 从",fromdate,"至",todate)
        return None, None, None     

    return True, start, end

if __name__ =="__main__":
    check_period('2020-1-1','2020-2-4')
    check_period('2020-1-1','2010-2-4')

#==============================================================================

#==============================================================================
#==============================================================================

#==============================================================================
#==============================================================================
